#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		Printer Server Process for the HP LaserJet Printer
 *                              (HPLJSERVER)
 *
 *		       Actually a modified QUMESERVER 
 *				Modified by
 *				J. W. Kinch
 *				1 March 1985
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>
#include "queue.h"
#include "qreplys.h"

time_t	timenow;
struct	request r;
int	lpoptions;
long	linecount = 0L;	/* number of lines printed by this server */
char	source[LINESIZ]; /* source of request, account name or net address */
char	obuf[BUFSIZ];	/* remember the output goes to stdout */
char	*arg0;		/* Pointer to the original argv[0] */
char	*title = 0;

#if BSD42 | BSD41c | V7
#include <sgtty.h>
struct sgttyb ttyb =	/* tty line setup commands */
{
	B9600,B9600,
	0,0,
	ANYP|LITOUT
};
#endif
#if BRLUNIX
#include <sgtty.h>
struct sg_brl ttyb =	/* tty line setup commands */
{
	B9600,B9600,
	DC3DC1|STALL,
	ANYP
};
#endif

struct	request *gethdr();
char	*index();
char	*strdup();
int	doflush();
int	dorestart();

extern	char *Qdatadir;

#define	FF	014	/* A formfeed character */

int	pg;		/* Used to save page size */
int	or;		/* Used to save page orientation */
int	ptch;		/* Used to save pitch (used to set margin) */
/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	datafile[LINESIZ];
	char	linebuf[LINESIZ];
	char	msg[LINESIZ];
	char	*cp;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	arginit(argc, argv);
	time( &timenow );
	setbuf( stdout, obuf );


	if (ioctl (fileno( stdout), TIOCSETP,  &ttyb) < 0) {
		err( "%s: Error setting HPLJ mode (errno %d)",
			arg0, errno);
		retmsg (RP_FATAL, "Error setting HPLJ mode (errno %d)", errno);
		exit (10);
	}

	if (seteightbits (fileno( stdout)) < 0) {
		err( "%s: Error setting HPLJ LITOUT mode (errno %d)",
			arg0, errno);
		retmsg (RP_FATAL, "Error setting HPLJ LITOUT mode (errno %d)", errno);
		exit (10);
	}


	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err( "%s: Error reading request", arg0);
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

		or = 0;			/* initialize flags for x options */
		pg = 0;
		ptch=0;
		printf ("\033E");	/* Reset printer */
		while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'D':
			if( chdir( &linebuf[1] ) < 0 ) {
				printf("\n\nError trying chdir (%s,%s)\n%c",
				&linebuf[1], sys_errlist[errno], FF );
				chdir( "/tmp" );	/* A safe default */
			}
			break;

		case 'O':
			setoptions (&linebuf[1]);
			break;

		case 'I':
			sprintf (msg, "printing %s", title);
			statmsg (msg);
			sprintf (datafile, "%s/%s", Qdatadir, &linebuf[1]);
			print (datafile);
			unlink (datafile);
			break;

		case 'F':
			sprintf (msg, "printing %s", &linebuf[1]);
			statmsg (msg);
			if (print (&linebuf[1]) < 0)
				break;
			if (lpoptions & NOKEEP)
				unlink (&linebuf[1]);
			break;

		case 'T':
			if (title) free(title);
			title = strdup(&linebuf[1]);
			break;

		case 'X':
			doxarg (&linebuf[1]);
			break;

		case 'B':
		case 'H':
		case 'U':
		default:
			/* Ignore other lines */
			break;
		}
	}
	retmsg (RP_OK, 0);
	exit (0);
}

/*
 *		S E T O P T I O N S  ( )
 *  Setoptions is passed the ascii string numerical representation
 *  of the lpoptions flags word.  The bit definitions are in queue.h.
 *  The string is a decimal representation.
 */
setoptions (optstr)
char *optstr;
{
	lpoptions = atoi (optstr);
}

/*
 *		D O X A R G  ( )
 *  Parse and act on extended arguments.
 */
doxarg (strp)
register char *strp;
{
	if ((strcmp (strp, "pg11"))==0)
	{
		if (pg != 0)	/* This is an error */
			err( "%s: Multiple Page Options", arg0);
		pg=11;		/* set flag that page was set */
		/* set 11-inch page option */
		/* at the same time set default line termination
		   and enable enf-of-line wrap */
		printf ("\033&l66P\033&k2G\033&s0C");
	}
	else if ((strcmp (strp, "pg14"))==0)
	{
		if (pg != 0)	/* This is an error */
			err( "%s: Multiple Page Options", arg0);
		pg=14;		/* set flag that page was set */
		/* set 14-inch page option */
		printf ("\033&l84P\033&k2G\033&s0C");
	}
	else if ((strcmp (strp, "por"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}
		printf ("\033&l0O");	/* set for portrait mode */
		or = 0;
	}
	else if ((strcmp (strp, "land"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}
		printf ("\033&l1O");	/* set for landscape mode */
		or = 1;
	}
	else if ((strcmp (strp, "manfeed"))==0)
	/* Selects the manual feed option on the printer */
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}
		printf ("\033&l2H");	/* Select manual feed */
	}
	else if ((strcmp (strp, "fntc10"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}

		/* select 10 pitch courier font
			ESC(8U selects the Roman-8 Symbol Set
			ESC(s0P selects fixed spacing
			ESC(s10H selects the 10 pitch font
			ESC(s12V selects the point size (12)
			ESC(s0S selects upright style
			ESC(s0B selects a medium stroke weight
			ESC(s4T selects the Courier type face
			Note: the ESC(s sequences are combined.
		Since the printer will select the font closest to that
		requested, it is the users responsibility to ensure that
		any required font cartridges are loaded.  He may do this
		with the -f option on qpr.  Also set the line spacing to
		six per inch in case someone reset it.
		*/

		printf ("\033&l6D\033(8U\033(s0p10h12v0s0b4T");
		ptch = 10;
	}
	else if ((strcmp (strp, "fnte12"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}

		/* selects the 12 picth elite font.  See fnt10 above */

		printf ("\033&l6D\033(8U\033(s0p12h10v0s0b3T");
		ptch = 12;
	}
	else if ((strcmp (strp, "fntlpr16"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}

		/* selects the 16.66 pitch line printer font.
		   See fnt10 above. Also sets the line spacing
	           to 8 line per inch.
		*/

		printf ("\033&l8D\033(8U\033(s0p16h8v0s0b0T");
		ptch = 16;
	}
	else if ((strcmp (strp, "nroff"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P");
		}
		/* resets the right, left and top margin to zero,
	 	   the page lenght to
		   either 66, 84 or 51 depending on page size and
		   orientation, set the line termination to zero,
	           and disables end-of-line wrap. */

		printf ("\0339\033&l0E");

		if (( pg == 11) && (or == 0))
			printf ("\033&l66F");
		else if (( pg == 14) && (or == 0))
			printf ("\033&l84F");
		else
			printf ("\033&l51F");

		printf ("\033&k0G\033&s1C");
	}
	else if ((strcmp (strp, "marg"))==0)
	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}

		/* set the left margin  to approximately 3/4 inch
		 *	7 for 10 pitch
		 *	9 for 12 pitch
		 *	12 for 16 ptich
		 */
		if ((ptch == 0) || (ptch == 10))
			printf ("\033&a7L");
		if (ptch == 12)
			printf ("\033&a9L");
		if (ptch == 16)
			printf ("\033&a12L");
	}


	/* This default is the same as por */
	else

	{
		if (pg == 0)	/* If page lenght hasn't been set */
		{		/* set to 11-inches */
			pg=11;	/* set flag that page was set */
			printf ("\033&l66P\033&k2G\033&s0C");
		}
		printf ("\033&l0O");	/* set for portrait mode */
		or = 0;
	}
}

/*
 *		P R I N T  ( )
 *  Dump a file onto the printer.  We do no checking
 *  as were are likely to be sending binary printer codes.
 */
print( file )
char *file;
{
	FILE	*df;
	int	cnt;

	if ((df = fopen (file, "r")) == NULL)
	{
		printf ("\n\n***** Could not open input file (%s, %s) *****\n",
			file, sys_errlist[errno] );
		putchar( FF );
		return( -1 );
	}
	for( cnt = r.r_copies; cnt > 0; cnt--)
	{
		if (copy (df, stdout) < 0)
			return (-1);
		if( cnt > 1 )
		{
			putchar ( FF );
			rewind( df );
		}
		else 	printf ( "\033E" );
	}
	fclose( df );
	return( 0 );
}
/*
 *		C O P Y ( )
 *  Takes two file pointers, and copies from one to the other
 */
copy (from, to)
FILE *from, *to;
{
	char	buf[BUFSIZ];
	int	nread;

	while ((nread = fread (buf, 1, BUFSIZ, from)) > NULL) {
		fwrite (buf, 1, nread, to);
		if (ferror(to)) {
			err ("Error reading source file (%s)",
				sys_errlist[errno]);
			return (-1);
		}
		if (ferror(from)) {
			err ("Error writing to printer (%s)",
				sys_errlist[errno]);
			return (-1);
		}
	}
	return (0);
}
/*
 *	A R G I N I T
 *
 *	Parse args of lineprinter like servers
 */
arginit(argc, argv)
int	argc;
char	**argv;
{
	int	baud;
	char	*parp;

	arg0 = argv[0];

	while (*++argv) {
 		if (strncmp (*argv, "speed=", 6) == SAME) {

      			if ((baud = atoi (*argv+6)) == 0) {
				err ("Bad argument '%s'", argv[1]);
				continue;
			}
			switch (baud) {
			case 134:   ttyb.sg_ispeed = ttyb.sg_ospeed = B134; break;
			case 150:   ttyb.sg_ispeed = ttyb.sg_ospeed = B150; break;
			case 300:   ttyb.sg_ispeed = ttyb.sg_ospeed = B300; break;
			case 600:   ttyb.sg_ispeed = ttyb.sg_ospeed = B600; break;
			case 1200:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1200; break;
			case 1800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1800; break;
			case 2400:  ttyb.sg_ispeed = ttyb.sg_ospeed = B2400; break;
			case 4800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B4800; break;
			case 9600:  ttyb.sg_ispeed = ttyb.sg_ospeed = B9600; break;
			default:
				err("Unknown speed %d", baud);
				break;
			}
			continue;
		}
		if (strncmp (*argv, "parity=", 7) == SAME) {
			 parp = *argv+7;
			if (strncmp (parp, "odd", 3) == SAME) {
				ttyb.sg_flags = TANDEM|CRMOD|XTABS|ODDP;
			 } else if (strncmp (parp, "even", 4) == SAME) {
				ttyb.sg_flags = TANDEM|CRMOD|XTABS|EVENP;
#ifdef SEL
			 } else if (strncmp (parp, "mark", 4) == SAME) {
			 	ttyb.sg_flags = TANDEM|CRMOD|XTABS|MARKP;
			 } else if (strncmp (parp, "space", 4) == SAME) {
			 	ttyb.sg_flags = TANDEM|CRMOD|XTABS|SPACEP;
#endif
			} else {
				err("Unknown parity '%s'", parp);
				continue;
			}
		}	
	}
}

/*
 *		E R R ( )
 *	Called by banner() and its subfunctions, amongst others
 */
/* VARARGS */
err( fmt, a, b, c, d )
char *fmt;
{
	fprintf( stderr, "%s: ", arg0);
	fprintf( stderr, fmt, a, b, c, d );
	putc( '\n', stderr );
}

seteightbits(ttyfd)
int ttyfd;
{
	int bits;
	struct sgttyb sg;

	bits = LLITOUT;
	if (ioctl(ttyfd, TIOCLBIS, &bits))
		return (-1);	/* something went wrong */
	/* theoretically we're done now, but most 4BSD drivers have a
	   small bug, which we get around as follows: */
	if (ioctl(ttyfd, TIOCGETP, &sg) || ioctl(ttyfd, TIOCSETN, &sg))
		return (-1);
	return (0);		/* success */
}

/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 *  This version is for the HP LaserJet server only
 */

dorestart()
{
	printf ("\033E");	/* reset the printer to force out page */
	retmsg (RP_RESTART, "RESTART requested");
	exit( 0 );
}


/*
 *	DOFLUSH ()
 *
 *  This version is for the HP LaserJet server only
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	printf ("\033E");	/* reset the printer to force out page */
	retmsg (RP_FLUSH, "Server received FLUSH request");
	exit( 0 );
}
